from PIL import Image
from enum import Enum

class Maze:
  Direction = Enum('Direction', [('NORTH', 1), ('EAST', 2), ('SOUTH', 3), ('WEST', 4)])
  
  start_color = (255,0,0)
  exit_color = (0,255,0)
  marker_color = (255,165,0)
  path_color = (0,0,255)
  
  def __init__(self,path):
    self.i = 0
    self.maze = Image.open(path, 'r')
    self.pixels = self.maze.load()
    self.size = self.maze.size
    self.direction = self.Direction.SOUTH
    self.position = self.find_start()
    if self.position is None:
      raise Exception(f"Can't find start position. Please add one pixel with color {self.start_color} to your image!")
    print("Loading done. Let's go!")
  
  def get_next_position(self):
    match self.direction:
      case self.Direction.NORTH:
        return (self.position[0], self.position[1] - 1)
      case self.Direction.EAST:
        return (self.position[0] + 1, self.position[1])
      case self.Direction.SOUTH:
        return (self.position[0], self.position[1] + 1)
      case self.Direction.WEST:
        return (self.position[0] - 1, self.position[1])
  
  def get_adjacent_positions(self, position=None):
    if position is None:
      position = self.position
    return [(self.Direction.NORTH,(position[0], position[1] - 1)),
            (self.Direction.EAST,(position[0] + 1, position[1])),
            (self.Direction.SOUTH,(position[0], position[1] + 1)),
            (self.Direction.WEST,(position[0] - 1, position[1]))]
  
  def turn_left(self):
    match self.direction:
      case self.Direction.NORTH:
        self.direction = self.Direction.WEST
      case self.Direction.EAST:
        self.direction = self.Direction.NORTH
      case self.Direction.SOUTH:
        self.direction = self.Direction.EAST
      case self.Direction.WEST:
        self.direction = self.Direction.SOUTH
  
  def turn_right(self):
    match self.direction:
      case self.Direction.NORTH:
        self.direction = self.Direction.EAST
      case self.Direction.EAST:
        self.direction = self.Direction.SOUTH
      case self.Direction.SOUTH:
        self.direction = self.Direction.WEST
      case self.Direction.WEST:
        self.direction = self.Direction.NORTH
  
  def can_move(self, pos = None):
    if pos is None:
      pos = self.get_next_position()
    x,y = pos
    if x < 0 or y < 0 or x >= self.size[0] or y >= self.size[1]:
      return False
    new_col = self.pixels[x, y]
    if new_col != (0, 0, 0) and new_col != self.start_color:
      return True
    return False
  
  def is_exit(self, position=None):
    if position is None:
      position = self.position
    x,y = position
    if self.pixels[x,y] == self.exit_color:
      print(f"Congratulations! You found the exit with {self.i} moves!")
      return True
    else:
      return False
  
  def move(self):
    self.position = self.get_next_position()
    self.maze.putpixel(self.position, self.path_color)
    self.i += 1
  
  def find_start(self):
    for x in range(self.size[0]):
      for y in range(self.size[1]):
        if self.pixels[x,y] == self.start_color:
          print(f"Found start position at {x},{y}. Let's go!")
          return ( x , y )
    return None
  
  def save_status(self, use_i_in_filename=False):
    if use_i_in_filename:
      self.maze.save(f"maze_{self.i}.png")
    else:
      self.maze.save(f"maze_path.png")

  def jump_to(self, position):
    print(f"jump from {self.position} to {position}")
    self.position = position

  def mark_position(self):
    self.maze.putpixel(self.position, self.marker_color)
    return (self.position, self.direction)

  def position_marked(self, position = None):
    if position is None:
      position = self.position
    return self.pixels[position[0], position[1]] == self.marker_color
  
  def position_visited_before(self, position = None):
    if position is None:
      position = self.position
    return self.pixels[position[0], position[1]] == self.path_color
